/* Linear hex pattern finder */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#define INCL_DOSFILEMGR
#define INCL_DOSPROCESS
#define INCL_DOSERRORS
#include <os2.h>

#define HF_STDIN 0
#define BUFFERSIZE 49152  /* 48 kB */

short int     rec_flag = FALSE, verbose = FALSE, p_length, p_signif,
              shift[256], jump[256], jumpm, p_order[255];
unsigned long filecount = 0L, bytecount = 0L, matchcount = 0L,
              fileopt = FILE_ARCHIVED | FILE_READONLY;
unsigned char buffer[BUFFERSIZE + 511], pattern[256],
              ch_prob [256] =
{
  0,   3,   7,   9,   8,  20,  24,  18,  12,  48,  15,  57,  41,  31,  54,  37,
 25,  72, 104, 121,  71, 107,  99, 144,  90, 149, 151, 157,  98, 156, 110, 131,
  2, 142,  80, 150,  33, 136,  59, 162,  92, 114, 113, 127,  75,  82,  69, 108,
 23,  47,  60,  44,  79,  89,  85, 109,  88,  84, 128, 118, 122, 106, 134, 138,
 46,  43,  93,  66,  35,  14,  34,  96,  97,  78, 180, 171,  68,  55,  87,  83,
 17, 126,  65,  50,  56,  73,  70,  94, 147, 177, 178, 169, 129, 141, 100,  63,
123,  16,  95,  49,  39,   4,  28,  74,  42,  13,  77, 146,  30,  61,  10,  19,
 45, 197,  11,  21,   6,  27,  62,  29, 111, 137, 168, 206, 135, 152, 116, 133,
 26, 101, 140,  22, 115,  91, 154, 174,  40,  36, 143,   5, 167,  51, 139, 189,
105, 182, 223, 193, 192, 237, 207, 241, 186, 155, 112, 196, 205, 248, 235, 229,
165, 161, 238, 212, 208, 221, 242, 250, 199, 253, 191, 255, 217, 254, 239, 243,
163, 234, 236, 252, 204, 247, 213, 233, 120, 200, 203, 124, 220, 246, 210, 164,
 32,  76, 132, 119,  52, 245, 170,  86, 172, 173, 201, 222,  58, 224, 211, 226,
159, 194, 184, 228, 215, 251, 230, 188, 160, 225, 232, 198, 185, 187, 240, 214,
130, 202, 216, 227, 175, 218, 231, 244,  38, 102, 219, 125, 103, 249, 166, 190,
 67, 181, 179, 195, 153, 209, 148, 145,  81, 183, 117, 176,  53, 158,  64,   1
}; /* ordinal number of decreasing character probability */


char nibble (char c)
{
  c = tolower (c);
  if (c < '0' || c > '9' && c < 'a' || c > 'f')
  {
    fputs ("Error in hexadecimal pattern.\n", stderr);
    DosExit (EXIT_PROCESS, 2);
  }

  return c < 'a' ? c - '0' : c - ('a' - 10);
} /* nibble */


void ParseOption (char *option)
{
  do
  {
    switch (tolower (*option))
    {
      case 'h':
        switch (*++option)
        {
          case '-':
            option++;
            fileopt &= ~FILE_HIDDEN;
            break;
          case '+':
            option++;
          default:
            fileopt |= FILE_HIDDEN;
        } /* endswitch h */
        break;

      case 's':
        switch (*++option)
        {
          case '-':
            option++;
            fileopt &= ~FILE_SYSTEM;
            break;
          case '+':
            option++;
          default:
            fileopt |= FILE_SYSTEM;
        } /* endswitch s */
        break;

      case 'r':
        switch (*++option)
        {
          case '-':
            option++;
            rec_flag = FALSE;
            break;
          case '+':
            option++;
          default:
            rec_flag = TRUE;
        } /* endswitch r */
        break;

      case 'v':
        switch (*++option)
        {
          case '-':
            option++;
            verbose = FALSE;
            break;
          case '+':
            option++;
          default:
            verbose = TRUE;
        } /* endswitch r */
        break;

      default:
        fprintf (stderr, "Unrecognized option \"%s\".\n", option);
        DosExit (EXIT_PROCESS, 3);

    } /* endswitch */
  } while (*option);

} /* ParseOption */


void ParsePattern (char *hexarg)
{
  register int i, j;
  int state, lastw;
  short int wild[256];

  p_length = 0;
  p_signif = 0;
  lastw = 0;
  state = 0;

  do
  {
    switch (state)
    {
      case 0:
        switch (*hexarg)
        {
          case '"':        /* double quoted string */
          case '\'':       /* single quoted string */
            state = *hexarg++;
            break;
          case ' ':
            hexarg++;      /* ignore blanks */
            break;
          case '?':        /* wildcard */
            wild[p_length] = TRUE;
            pattern[p_length++] = *hexarg++;
            lastw = p_length;
            break;
          default:
            wild[p_length] = FALSE;
            p_order[p_signif++] = p_length;
            pattern[p_length] = nibble (*hexarg++) << 4;
            state = 1;
        }
        break;
      case 1:
        pattern[p_length++] |= nibble (*hexarg++);
        state = 0;
        break;
      default:
        if (*hexarg == state)
        {
          hexarg++;
          state = 0;
        }
        else
        {
          wild[p_length] = FALSE;
          p_order[p_signif++] = p_length;
          pattern[p_length++] = *hexarg++;
        }
    } /* endswitch state */

  } while (*hexarg && p_length < 256); /* enddo */

  if (lastw == p_length)
  {
    fputs ("Trailing wildcard not allowed.\n", stderr);
    DosExit (EXIT_PROCESS, 4);
  }

  /* insertion sort p_order up to p_signif - 2 in ch_prob order */
  for (i = 1; i < p_signif - 1; i++)
  {
    state = p_order[i];
    for (j = i; j > 0; j--)
    {
      if (ch_prob[pattern[p_order[j - 1]]] > ch_prob[pattern[state]])
        p_order[j] = p_order[j - 1];
      else
        break;
    }
    p_order[j] = state;
  }

  /* make shift and jump tables */
  for (i = 0; i < 256; i++)
     shift[i] = p_length - lastw;
  for (i = 0; i < p_signif; i++)
  {
    if (p_order[i] >= lastw)
      shift[pattern[p_order[i]]] = p_length - p_order[i] - 1;
    jump[i] = p_length;
  }
  jumpm = p_length;
  for (j = p_length - 1; j > 0; j--)
  {
    for (i = p_signif - 1; i >= 0; i--)
    {
      if (p_order[i] < j || wild[p_order[i] - j])
        jump[i] = j;
      else if (pattern[p_order[i]] != pattern[p_order[i]-j])
      {
        jump[i] = j;
        break;
      }
    }
    if (i < 0)
      jumpm = j;
  }

  printf ("Searching for HEX pattern ");
  for (i = 0; i < p_length; i++)
    if (wild[i])
      printf ("?? ");
    else
      printf ("%02X ", pattern[i]);
  printf ("...\n");

} /* ParsePattern */


unsigned char *FastSearch (register unsigned char *where)
{
  register int i, plm1;

  plm1 = p_length - 1;
  for ( ; ; )
  {
    while (i = shift[where[plm1]])
      where += i;

    i = p_signif - 1;
    do
      if (--i < 0)
        return where;
    while (where[p_order[i]] == pattern[p_order[i]]);

    where += jump[i];
  } /* inner loop */

} /* FastSearch */


void SearchFile (char *filename, HFILE handle)
{
  ULONG result, f_offset;
  unsigned char *fill, *found, pdc[12];

  if (verbose)
    printf ("Searching %s ...\n", filename);

  f_offset = 0;
  fill = buffer;
  pdc[0] = '(';

  while (DosRead (handle, fill, BUFFERSIZE, &result) == NO_ERROR
         && result)
  {
    /* statistics */
    bytecount += result;

    /* append pattern after end */
    fill += result;
    memcpy (fill, pattern, p_length);

    found = FastSearch (buffer);

    while (found <= fill - p_length)
    {
      /* it's a real match - print and count it */
      ltoa ((found - buffer) + f_offset, &pdc[1], 10);
      printf ("offset 0x%08lX%11s) in %s\n",
              (found - buffer) + f_offset, pdc, filename);
      matchcount++;

      found = FastSearch (found + jumpm);
    }

    if (result < BUFFERSIZE)
    {
      break;
    }
    else
    {
      result = p_length - 1;
      memcpy (buffer, fill - result, result);
      f_offset += fill - buffer - result;
      fill = buffer + result;
    }
  }

  /* statistics */
  filecount++;

} /* SearchFile */


int SearchPath (char *path, char *mask)
{
  int rc;
  ULONG count, result;
  HDIR dhandle;
  HFILE fhandle;
  FILEFINDBUF3 finfo;
  char FullPathName[CCHMAXPATH];

  /* 1. search files */
  strcpy (FullPathName, path);
  strcat (FullPathName, mask);

  dhandle = HDIR_CREATE;
  count = 1;

  if (DosFindFirst (FullPathName,
                    &dhandle,
                    fileopt,
                    &finfo,
                    sizeof (finfo),
                    &count,
                    FIL_STANDARD) == NO_ERROR)
  {
    do
    {
      strcpy (FullPathName, path);
      strcat (FullPathName, finfo.achName);

      if (DosOpen (FullPathName,
                   &fhandle,
                   &result,
                   0L,
                   FILE_NORMAL,
                   OPEN_ACTION_OPEN_IF_EXISTS,
                   OPEN_FLAGS_SEQUENTIAL |
                     OPEN_SHARE_DENYNONE |
                     OPEN_ACCESS_READONLY,
                   0L))
      {
        fprintf (stderr, "Error opening file %s.\n", FullPathName);
      }
      else
      {
        SearchFile (FullPathName, fhandle);
        DosClose (fhandle);
      }
    } while (DosFindNext (dhandle,
                          &finfo,
                          sizeof (finfo),
                          &count) == NO_ERROR); /* enddo */
    DosFindClose (dhandle);
    rc = 0;
  }
  else
  {
    rc = 1;  /* no files found */
  } /* endif DosFindFirst */

  /* 2. if recursive search subdirectories */
  if (rec_flag)
  {
    strcpy (FullPathName, path);
    strcat (FullPathName, "*");

    dhandle = HDIR_CREATE;
    count = 1;

    if (DosFindFirst (FullPathName,
                      &dhandle,
                      fileopt | MUST_HAVE_DIRECTORY,
                      &finfo,
                      sizeof (FILEFINDBUF3),
                      &count,
                      FIL_STANDARD) == NO_ERROR)
    {
      do
      {
        if (strcmp (finfo.achName, ".") && strcmp (finfo.achName, ".."))
        {
          strcpy (FullPathName, path);
          strcat (FullPathName, finfo.achName);
          strcat (FullPathName, "\\");

          rc &= SearchPath (FullPathName, mask);
        }
      } while (DosFindNext (dhandle,
                            &finfo,
                            sizeof (finfo),
                            &count) == NO_ERROR); /* enddo */
      DosFindClose (dhandle);
    } /* endif DosFindFirst */
  } /* endif */

  return rc;
} /* SearchPath */


int main (int argc, char *argv[])
{
  int i;
  char rootpath[CCHMAXPATH], filemask[CCHMAXPATHCOMP], *filename = NULL;

  printf ("HexFind 2.0 (c) 1998/99 by Heinz Repp\n");

  /* parse command line: */

  /* 1. leading options */
  i = 1;
  while (i < argc && (*argv[i] == '/' || *argv[i] == '-'))
    ParseOption (++argv[i++]);

  /* 2. pattern */
  if (i < argc)
  {
    ParsePattern (argv[i]);
  }
  else
  {
    fputs ("Usage: HexFind [-options] [\"]<pattern>[\"] files ...\n"
           "       options: <r>ecursive, <v>erbose, <h>idden or <s>ystem files\n"
           "       <pattern>: one or more of [ HEX-pair | 'string' | ? ]\n"
           "       (? = wildcard, pattern must be contiguous or quoted!)\n",
           stderr);
    DosExit (EXIT_PROCESS, 1);
  }

  /* 3. trailing options and file specs */
  while (++i < argc)
  {
    if (*argv[i] == '/' || *argv[i] == '-')
      ParseOption (++argv[i]);
    else
    {
      strcpy (rootpath, argv[i]);
      if ((filename = strrchr (rootpath, '\\')) ||
          (filename = strrchr (rootpath, '/')))
        filename++;   /* skip path delimiter */
      else
        filename = rootpath;      /* no path */
      strcpy (filemask, filename);
      *filename = '\0';

      if (SearchPath (rootpath, filemask))
        fprintf (stderr, "Filespec '%s%s' not found.\n", rootpath, filemask);
    } /* endif */
  } /* endwhile all arguments */

  /* if no file spec until now */
  if (filename == NULL)
    SearchFile ("StdIn", HF_STDIN);

  if (verbose)
    printf ("Found %lu occurence(s) in %lu file(s) (%lu bytes searched).\n",
             matchcount, filecount, bytecount);

  DosExit (EXIT_PROCESS, 0);
} /* main */
